@@ -14,7 +14,7 @@ along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 **********/
 // "liveMedia"
-// Copyright (c) 1996-2024 Live Networks, Inc.  All rights reserved.
+// Copyright (c) 1996-2025 Live Networks, Inc.  All rights reserved.
 // A sink that generates a QuickTime file from a composite media session
 // Implementation
 
@@ -26,13 +26,15 @@ along with this library; if not, write to the Free Software Foundation, Inc.,
 #include "H263plusVideoRTPSource.hh" // for the special header
 #include "MPEG4GenericRTPSource.hh" //for "samplingFrequencyFromAudioSpecificConfig()"
 #include "MPEG4LATMAudioRTPSource.hh" // for "parseGeneralConfigStr()"
+#include "H264or5VideoStreamFramer.hh" // for "removeH264or5EmulationBytes()"
 #include "Base64.hh"
 
 #include <ctype.h>
 
 #define fourChar(x,y,z,w) ( ((x)<<24)|((y)<<16)|((z)<<8)|(w) )
 
-#define H264_IDR_FRAME 0x65  //bit 8 == 0, bits 7-6 (ref) == 3, bits 5-0 (type) == 5
+#define H264_IDR_FRAME 0x65  //bit 8 == 0, bits 7-6 (ref) == 3, bits 5-1 (type) == 5
+#define isIDRFrame(firstByte) (firstByte == H264_IDR_FRAME || ((firstByte&0x7E)>>1) == 19 || ((firstByte&0x7E)>>1) == 20)
 
 ////////// SubsessionIOState, ChunkDescriptor ///////////
 // A structure used to represent the I/O state of each input 'subsession':
@@ -653,6 +655,10 @@ Boolean SubsessionIOState::setQTstate() {
 	fQTMediaDataAtomCreator = &QuickTimeFileSink::addAtom_avc1;
 	fQTTimeScale = 600;
 	fQTTimeUnitsPerSample = fQTTimeScale/fOurSink.fMovieFPS;
+      } else if (strcmp(fOurSubsession.codecName(), "H265") == 0) {
+	fQTMediaDataAtomCreator = &QuickTimeFileSink::addAtom_hvc1;
+	fQTTimeScale = 600;
+	fQTTimeUnitsPerSample = fQTTimeScale/fOurSink.fMovieFPS;
       } else if (strcmp(fOurSubsession.codecName(), "MP4V-ES") == 0) {
 	fQTMediaDataAtomCreator = &QuickTimeFileSink::addAtom_mp4v;
 	fQTTimeScale = 600;
@@ -801,7 +807,10 @@ void SubsessionIOState::useFrame(SubsessionBuffer& buffer) {
   struct timeval const& presentationTime = buffer.presentationTime();
   int64_t const destFileOffset = TellFile64(fOurSink.fOutFid);
   unsigned sampleNumberOfFrameStart = fQTTotNumSamples + 1;
-  Boolean avcHack = fQTMediaDataAtomCreator == &QuickTimeFileSink::addAtom_avc1;
+  Boolean h264or5Hack =
+    fQTMediaDataAtomCreator == &QuickTimeFileSink::addAtom_avc1 ||
+    fQTMediaDataAtomCreator == &QuickTimeFileSink::addAtom_hvc1;
+    
 
   // If we're not syncing streams, or this subsession is not video, then
   // just give this frame a fixed duration:
@@ -809,7 +818,7 @@ void SubsessionIOState::useFrame(SubsessionBuffer& buffer) {
       || fQTcomponentSubtype != fourChar('v','i','d','e')) {
     unsigned const frameDuration = fQTTimeUnitsPerSample*fQTSamplesPerFrame;
     unsigned frameSizeToUse = frameSize;
-    if (avcHack) frameSizeToUse += 4; // H.264/AVC gets the frame size prefix
+    if (h264or5Hack) frameSizeToUse += 4; // H.264/5 gets the frame size prefix
 
     fQTTotNumSamples += useFrame1(frameSizeToUse, presentationTime, frameDuration, destFileOffset);
   } else {
@@ -825,7 +834,7 @@ void SubsessionIOState::useFrame(SubsessionBuffer& buffer) {
       unsigned frameDuration
 	= (unsigned)((2*duration*fQTTimeScale+1)/2); // round
       unsigned frameSizeToUse = fPrevFrameState.frameSize;
-      if (avcHack) frameSizeToUse += 4; // H.264/AVC gets the frame size prefix
+      if (h264or5Hack) frameSizeToUse += 4; // H.264/5 gets the frame size prefix
 
       unsigned numSamples
 	= useFrame1(frameSizeToUse, ppt, frameDuration, fPrevFrameState.destFileOffset);
@@ -833,7 +842,7 @@ void SubsessionIOState::useFrame(SubsessionBuffer& buffer) {
       sampleNumberOfFrameStart = fQTTotNumSamples + 1;
     }
 
-    if (avcHack && (*frameSource == H264_IDR_FRAME)) {
+    if (h264or5Hack && isIDRFrame(*frameSource)) {
       SyncFrame* newSyncFrame = new SyncFrame(fQTTotNumSamples + 1);
       if (fTailSyncFrame == NULL) {
         fHeadSyncFrame = newSyncFrame;
@@ -849,7 +858,7 @@ void SubsessionIOState::useFrame(SubsessionBuffer& buffer) {
     fPrevFrameState.destFileOffset = destFileOffset;
   }
 
-  if (avcHack) fOurSink.addWord(frameSize);
+  if (h264or5Hack) fOurSink.addWord(frameSize);
 
   // Write the data into the file:
   fwrite(frameSource, 1, frameSize, fOurSink.fOutFid);
@@ -1119,7 +1128,7 @@ Boolean SubsessionIOState::syncOK(struct timeval presentationTime) {
 
 	  // if audio is in sync, wait for the next IDR frame to start
 	  unsigned char* const frameSource = fBuffer->dataStart();
-	  if (*frameSource != H264_IDR_FRAME) return False;
+	  if (!isIDRFrame(*frameSource)) return False;
 	}
 	// But now we are
 	fHaveBeenSynced = True;
@@ -1888,43 +1897,163 @@ addAtom(avc1);
 addAtomEnd;
 
 addAtom(avcC);
-// Begin by Base-64 decoding the "sprop" parameter sets strings:
+  // Begin by Base-64 decoding the "sprop" parameter sets strings:
   char* psets = strDup(fCurrentIOState->fOurSubsession.fmtp_spropparametersets());
   if (psets == NULL) return 0;
 
+  char const* spsBase64 = psets;
   size_t comma_pos = strcspn(psets, ",");
   psets[comma_pos] = '\0';
-  char const* sps_b64 = psets;
-  char const* pps_b64 = &psets[comma_pos+1];
-  unsigned sps_count;
-  unsigned char* sps_data = base64Decode(sps_b64, sps_count, false);
-  unsigned pps_count;
-  unsigned char* pps_data = base64Decode(pps_b64, pps_count, false);
 
-// Then add the decoded data:
+  char const* ppsBase64 = &psets[comma_pos+1];
+
+  unsigned spsSize;
+  unsigned char* sps = base64Decode(spsBase64, spsSize, false);
+
+  unsigned ppsSize;
+  unsigned char* pps = base64Decode(ppsBase64, ppsSize, false);
+
+  // We use some of the data from the "SPS". Remove any 'emulation bytes' from it first:
+  if (spsSize == 0) return 0;
+  u_int8_t* spsWEB = new u_int8_t[spsSize]; // "WEB" means "Without Emulation Bytes"
+  unsigned spsWEBSize = removeH264or5EmulationBytes(spsWEB, spsSize, sps, spsSize);
+  if (spsWEBSize < 4) { // Bad SPS size
+    delete[] spsWEB;
+    return 0;
+  }
+
+  // Then add the decoded data:
   size += addByte(0x01); // configuration version
-  size += addByte(sps_data[1]); // profile
-  size += addByte(sps_data[2]); // profile compat
-  size += addByte(sps_data[3]); // level
+  size += addByte(spsWEB[1]); // profile
+  size += addByte(spsWEB[2]); // profile compat
+  size += addByte(spsWEB[3]); // level
   size += addByte(0xff); /* 0b11111100 | lengthsize = 0x11 */
-  size += addByte(0xe0 | (sps_count > 0 ? 1 : 0) );
-  if (sps_count > 0) {
-    size += addHalfWord(sps_count);
-    for (unsigned i = 0; i < sps_count; i++) {
-      size += addByte(sps_data[i]);
+  size += addByte(0xe0 | (spsSize > 0 ? 1 : 0) );
+  if (spsSize > 0) {
+    size += addHalfWord(spsSize);
+    for (unsigned i = 0; i < spsSize; i++) {
+      size += addByte(sps[i]);
     }
   }
-  size += addByte(pps_count > 0 ? 1 : 0);
-  if (pps_count > 0) {
-    size += addHalfWord(pps_count);
-    for (unsigned i = 0; i < pps_count; i++) {
-      size += addByte(pps_data[i]);
+  size += addByte(ppsSize > 0 ? 1 : 0);
+  if (ppsSize > 0) {
+    size += addHalfWord(ppsSize);
+    for (unsigned i = 0; i < ppsSize; i++) {
+      size += addByte(pps[i]);
+    }
+  }
+
+  // Finally, delete the data that we allocated:
+  delete[] spsWEB;
+  delete[] pps; delete[] sps;
+  delete[] psets;
+addAtomEnd;
+
+addAtom(hvc1);
+// General sample description fields:
+  size += addWord(0x00000000); // Reserved
+  size += addWord(0x00000001); // Reserved+Data       reference index
+// Video sample       description     fields:
+  size += addWord(0x00000000); // Version+Revision level
+  size += add4ByteString("appl"); // Vendor
+  size += addWord(0x00000000); // Temporal quality
+  size += addWord(0x00000000); // Spatial quality
+  unsigned const widthAndHeight       = (fMovieWidth<<16)|fMovieHeight;
+  size += addWord(widthAndHeight); // Width+height
+  size += addWord(0x00480000); // Horizontal resolution
+  size += addWord(0x00480000); // Vertical resolution
+  size += addWord(0x00000000); // Data size
+  size += addWord(0x00010548); // Frame count+Compressor name (start)
+    // "H.265"
+  size += addWord(0x2e323635); // Compressor name (continued)
+  size += addZeroWords(6); // Compressor name (continued - zero)
+  size += addWord(0x00000018); // Compressor name (final)+Depth
+  size += addHalfWord(0xffff); // Color table id
+  size += addAtom_hvcC();
+addAtomEnd;
+
+addAtom(hvcC);
+  // Begin by Base-64 decoding the "sprop" parameter sets strings:
+  char const* vpsBase64 = strDup(fCurrentIOState->fOurSubsession.fmtp_spropvps());
+  char const* spsBase64 = strDup(fCurrentIOState->fOurSubsession.fmtp_spropsps());
+  char const* ppsBase64 = strDup(fCurrentIOState->fOurSubsession.fmtp_sproppps());
+  if (vpsBase64 == NULL || spsBase64 == NULL || ppsBase64 == NULL) return 0;
+
+  unsigned vpsSize;
+  unsigned char* vps = base64Decode(vpsBase64, vpsSize, false);
+
+  unsigned spsSize;
+  unsigned char* sps = base64Decode(spsBase64, spsSize, false);
+
+  unsigned ppsSize;
+  unsigned char* pps = base64Decode(ppsBase64, ppsSize, false);
+
+  // We use some of the data from the "VPS". Remove any 'emulation bytes' from it first:
+  if (vpsSize == 0) return 0;
+  u_int8_t* vpsWEB = new u_int8_t[vpsSize]; // "WEB" means "Without Emulation Bytes"
+  unsigned vpsWEBSize = removeH264or5EmulationBytes(vpsWEB, vpsSize, vps, vpsSize);
+  if (vpsWEBSize < 6/*'profile_tier_level' offset*/ + 12/*num 'profile_tier_level' bytes*/) {
+    // Bad VPS size
+    delete[] vpsWEB;
+    return 0;
+  }
+
+  // Then add the decoded data:
+  size += addByte(0x01); // configurationVersion = 1
+
+  u_int8_t const* profileTierLevelHeaderBytes = &vpsWEB[6];
+  size += addByte(profileTierLevelHeaderBytes[0]);
+      // general_profile_space + general_tier_flag + general_profile_idc
+
+  u_int8_t const* general_profile_compatibility_flags = &profileTierLevelHeaderBytes[1];
+  for (unsigned i = 0; i < 4; ++i) size += addByte(general_profile_compatibility_flags[i]);
+      // general_profile_compatibility_flags
+
+  u_int8_t const* general_constraint_indicator_flags = &profileTierLevelHeaderBytes[5];
+  for (unsigned i = 0; i < 6; ++i) size += addByte(general_constraint_indicator_flags[i]);
+      // general_constraint_indicator_flags
+
+  size += addByte(profileTierLevelHeaderBytes[11]); // general_level_idc
+  size += addHalfWord(0xF000); // min_spatial_segmentation_idc = 0 ???
+  size += addByte(0xFC); // parallelismType = 0 ???
+  size += addByte(0xFD); // chroma_format_idc = 1 ???
+  size += addByte(0xF8); // bit_depth_luma_minus8 = 0 ???
+  size += addByte(0xF8); // bit_depth_chroma_minus8 = 0 ???
+  size += addHalfWord(0x0000); // avgFrameRate = 0
+  size += addByte(0x0F);
+      // constantFrameRate = 0; numTemporalLayers = 1; temporalIdNested = 1; lengthSizeMinusOne = 3 ???
+  size += addByte(1 + (spsSize > 0) + (ppsSize > 0)); // numOfArrays
+
+  if (vpsSize > 0) {
+    size += addByte(0x20); // array_completeness = 0; NAL_unit_type = VPS
+    size += addHalfWord(1); // numNalus
+    size += addHalfWord(vpsSize);
+    for (unsigned i = 0; i < vpsSize; i++) {
+      size += addByte(vps[i]);
+    }
+  }
+
+  if (spsSize > 0) {
+    size += addByte(0x21); // array_completeness = 0; NAL_unit_type = SPS
+    size += addHalfWord(1); // numNalus
+    size += addHalfWord(spsSize);
+    for (unsigned i = 0; i < spsSize; i++) {
+      size += addByte(sps[i]);
+    }
+  }
+
+  if (ppsSize > 0) {
+    size += addByte(0x22); // array_completeness = 0; NAL_unit_type = PPS
+    size += addHalfWord(1); // numNalus
+    size += addHalfWord(ppsSize);
+    for (unsigned i = 0; i < ppsSize; i++) {
+      size += addByte(pps[i]);
     }
   }
 
 // Finally, delete the data that we allocated:
-  delete[] pps_data; delete[] sps_data;
-  delete[] psets;
+  delete[] vpsWEB;
+  delete[] vps; delete[] pps; delete[] sps;
 addAtomEnd;
 
 addAtom(mp4v);
